#!/usr/bin/env python3

import numpy as np
import os
import sys
import wfdb
import json
import joblib
from prepro import Preprocessing
from tensorflow.keras.models import load_model
from features import Features
from scipy.signal import medfilt
from sklearn.preprocessing import MinMaxScaler

"""
Written by:  Xingyao Wang, Chengyu Liu
             School of Instrument Science and Engineering
             Southeast University, China
             chengyu@seu.edu.cn


Modified by: Sen Liu, ya'nan Wang, Haijun Jia    


Save answers to '.json' files, the format is as {'predict_endpoints':[[s0, e0], [s1, e1], ... [sm-1, em-2]]}.
"""

def load_data(sample_path):
    sig, fields = wfdb.rdsamp(sample_path)
    sig = sig[:,1]
    fs = fields['fs']
    
    return sig, fs


def loadModels():
    
    model = load_model('model_vgg13_1043.h5')
    model_svm = joblib.load('model_svm.pkl')
    xtr = np.load('xtr.npy')
    
    return model, model_svm, xtr


def save_dict(filename, dic):
    '''save dict into json file'''
    with open(filename,'w') as json_file:
        json.dump(dic, json_file, ensure_ascii=False)
    
    

def challenge_entry(sample_path):
 
    
    sig, fs = load_data(sample_path)
    ecgB = Preprocessing(fs, sig, sample_path).data_filter()
    featureList = Features(sample_path)
    
    emp = []
    emp.append(featureList)
    featureList = np.array(emp)
    
    scaler = MinMaxScaler()
    scaler.fit(xtr)
    featureNorm = scaler.transform(featureList)
    pre = model_svm.predict(featureNorm)
    
    end_points =[]
    if pre ==  0:
        pred_dict = {'predict_endpoints': end_points}
        
   
    elif pre ==  1:
        end_points.append([0, len(ecgB)-1])
        pred_dict = {'predict_endpoints': end_points}
    
    
    else:
        x_test, beatPos_new, beatPos = Preprocessing(fs, ecgB, sample_path).beatread()
        x_test = np.array(x_test)
        x_test = x_test[:,:,np.newaxis]
        y_predict0 = model.predict(x_test)
        y_predict = y_predict0.squeeze(axis=1)
        ylist = [int(round(y)) for y in y_predict]  
        
        ylist = medfilt(np.array(ylist).squeeze(), 5)
        beatPos = np.array(beatPos)
        beatPos = beatPos[beatPos<len(sig)-1]
        num_sup = len(beatPos) - len(ylist)
        if num_sup == 1:
            if beatPos[0]<beatPos_new[0]:
                ylist = np.insert(ylist, 0, ylist[0])
            else:
                ylist = np.insert(ylist, len(ylist), ylist[-1])
        if num_sup > 1:
            ylist = np.insert(ylist, 0, ylist[0])
            for i in range(num_sup-1):
                ylist = np.insert(ylist, len(ylist), ylist[-1])
        
        state_diff = np.diff(ylist)
        start_r = np.where(state_diff==1)[0] + 1
        end_r = np.where(state_diff==-1)[0] + 1
        
        if ylist[0] == 1:
            start_r = np.insert(start_r, 0, 0)
        if ylist[-1] == 1:
            end_r = np.insert(end_r, len(end_r), len(ylist)-1)
        
        start_r = np.expand_dims(start_r, -1)
        end_r = np.expand_dims(end_r, -1)
        start_end = np.concatenate((beatPos[start_r], beatPos[end_r]), axis=-1).tolist()
        end_points.extend(start_end) 
        
        pred_dict = {'predict_endpoints': end_points}

    return pred_dict


if __name__ == '__main__':

    DATA_PATH = sys.argv[1]
    RESULT_PATH = sys.argv[2]
    
    if not os.path.exists(RESULT_PATH):
        os.makedirs(RESULT_PATH)
        
    test_set = open(os.path.join(DATA_PATH, 'RECORDS'), 'r').read().splitlines()
    model, model_svm, xtr = loadModels()
    
    for i, sample in enumerate(test_set):
        sample_path = os.path.join(DATA_PATH, sample)
        pred_dict = challenge_entry(sample_path)

        save_dict(os.path.join(RESULT_PATH, sample+'.json'), pred_dict)

