# -*- coding: utf-8 -*-
"""
Written by:  Sen Liu, ya'nan Wang, Haijun Jia
"""

from typing import List, Tuple
from collections import namedtuple
import numpy as np
from scipy import interpolate
from scipy import signal
from astropy.stats import LombScargle
import wfdb


# Frequency Methods name
WELCH_METHOD = "welch"
LOMB_METHOD = "lomb"

# Named Tuple for different frequency bands
VlfBand = namedtuple("Vlf_band", ["low", "high"])
LfBand = namedtuple("Lf_band", ["low", "high"])
HfBand = namedtuple("Hf_band", ["low", "high"])


def get_time_domain_features(nn_intervals: List[float], pnni_as_percent: bool = True) -> dict:
    ''''time domain features'''
    
    diff_nni = np.diff(nn_intervals, axis=0) 
    length_int = len(nn_intervals) - 1 if pnni_as_percent else len(nn_intervals)

    # Basic statistics
    mean_nni = np.mean(nn_intervals, axis=0)
    median_nni = np.median(nn_intervals, axis=0)
    range_nni = max(nn_intervals) - min(nn_intervals)

    sdsd = np.std(diff_nni, axis=0, ddof=1)
    rmssd = np.sqrt(np.mean(diff_nni ** 2))

    nni_50 = sum(np.abs(diff_nni) > 50)
    pnni_50 = 100 * nni_50 / length_int
    nni_20 = sum(np.abs(diff_nni) > 20)
    pnni_20 = 100 * nni_20 / length_int

    cvsd = rmssd / mean_nni

    # Features suitable for long term recordings
    sdnn = np.std(nn_intervals, ddof=1) 
    cvnni = sdnn / mean_nni

    # Heart Rate equivalent features
    heart_rate_list = np.divide(60000, nn_intervals)
    mean_hr = np.mean(heart_rate_list)
    min_hr = min(heart_rate_list)
    max_hr = max(heart_rate_list)
    std_hr = np.std(heart_rate_list)

    time_domain_features = {
        'mean_nni': mean_nni,
        'sdnn': sdnn,
        'sdsd': sdsd,
        'nni_50': nni_50,
        'pnni_50': pnni_50,
        'nni_20': nni_20,
        'pnni_20': pnni_20,
        'rmssd': rmssd,
        'median_nni': median_nni,
        'range_nni': range_nni,
        'cvsd': cvsd,
        'cvnni': cvnni,
        'mean_hr': mean_hr,
        "max_hr": max_hr,
        "min_hr": min_hr,
        "std_hr": std_hr,
    }

    return time_domain_features


def get_frequency_domain_features(nn_intervals: List[float], method: str = WELCH_METHOD,
                                  sampling_frequency: int = 7, interpolation_method: str = "linear",
                                  vlf_band: namedtuple = VlfBand(0.003, 0.04),
                                  lf_band: namedtuple = LfBand(0.04, 0.15),
                                  hf_band: namedtuple = HfBand(0.15, 0.40)) -> dict:
    
    '''frequency feature'''
    
    nn_intervals = list(nn_intervals)

    freq, psd = _get_freq_psd_from_nn_intervals(nn_intervals=nn_intervals, method=method,
                                                sampling_frequency=sampling_frequency,
                                                interpolation_method=interpolation_method,
                                                vlf_band=vlf_band, hf_band=hf_band)

    frequency_domain_features = _get_features_from_psd(freq=freq, psd=psd,
                                                      vlf_band=vlf_band,
                                                      lf_band=lf_band,
                                                      hf_band=hf_band)

    return frequency_domain_features


def _get_freq_psd_from_nn_intervals(nn_intervals: List[float], method: str = WELCH_METHOD,
                                    sampling_frequency: int = 7,
                                    interpolation_method: str = "linear",
                                    vlf_band: namedtuple = VlfBand(0.003, 0.04),
                                    hf_band: namedtuple = HfBand(0.15, 0.40)) -> Tuple:

    timestamp_list = _create_timestamp_list(nn_intervals)

    if method == WELCH_METHOD:
        # Interpolation of signal
        funct = interpolate.interp1d(x=timestamp_list, y=nn_intervals, kind=interpolation_method)

        timestamps_interpolation = _create_interpolated_timestamp_list(nn_intervals, sampling_frequency)
        nni_interpolation = funct(timestamps_interpolation)

        # Remove DC Component
        nni_normalized = nni_interpolation - np.mean(nni_interpolation)

        # Compute Power Spectral Density
        freq, psd = signal.welch(x=nni_normalized, fs=sampling_frequency, window='hann',
                                 nfft=4096)

    elif method == LOMB_METHOD:
        freq, psd = LombScargle(timestamp_list, nn_intervals,
                                normalization='psd').autopower(minimum_frequency=vlf_band[0],
                                                               maximum_frequency=hf_band[1])
    else:
        raise ValueError("Not a valid method. Choose between 'lomb' and 'welch'")

    return freq, psd


def _create_timestamp_list(nn_intervals: List[float]) -> List[float]:
    
    # Convert in seconds
    nni_tmstp = np.cumsum(nn_intervals, axis=0) / 1000 

    # Force to start at 0
    return nni_tmstp - nni_tmstp[0]


def _create_interpolated_timestamp_list(nn_intervals: List[float], sampling_frequency: int = 7) -> List[float]:

    time_nni = _create_timestamp_list(nn_intervals)
    # Create timestamp for interpolation
    nni_interpolation_tmstp = np.arange(0, time_nni[-1], 1 / float(sampling_frequency))
    return nni_interpolation_tmstp


def _get_features_from_psd(freq: List[float], psd: List[float], vlf_band: namedtuple = VlfBand(0.003, 0.04),
                           lf_band: namedtuple = LfBand(0.04, 0.15),
                           hf_band: namedtuple = HfBand(0.15, 0.40)) -> dict:
   
    # Calcul of indices between desired frequency bands
    vlf_indexes = np.logical_and(freq >= vlf_band[0], freq < vlf_band[1])
    lf_indexes = np.logical_and(freq >= lf_band[0], freq < lf_band[1])
    hf_indexes = np.logical_and(freq >= hf_band[0], freq < hf_band[1])

    # Integrate using the composite trapezoidal rule
    lf = np.trapz(y=psd[lf_indexes], x=freq[lf_indexes], axis=0)
    hf = np.trapz(y=psd[hf_indexes], x=freq[hf_indexes], axis=0)

    # total power & vlf 
    vlf = np.trapz(y=psd[vlf_indexes], x=freq[vlf_indexes], axis=0)
    total_power = vlf + lf + hf

    lf_hf_ratio = lf / hf
    lfnu = (lf / (lf + hf)) * 100
    hfnu = (hf / (lf + hf)) * 100

    freqency_domain_features = {
        'lf': lf,
        'hf': hf,
        'lf_hf_ratio': lf_hf_ratio,
        'lfnu': lfnu,
        'hfnu': hfnu,
        'total_power': total_power,
        'vlf': vlf
    }

    return freqency_domain_features


def get_csi_cvi_features(nn_intervals: List[float]) -> dict:

    # Measures the width and length of poincare cloud
    poincare_plot_features = get_poincare_plot_features(nn_intervals, Plot=False)
    T = 4 * poincare_plot_features['sd1']
    L = 4 * poincare_plot_features['sd2']

    csi = L / T
    cvi = np.log10(L * T)
    modified_csi = L ** 2 / T

    csi_cvi_features = {
        'csi': csi,
        'cvi': cvi,
        'Modified_csi': modified_csi
    }
        
    return csi_cvi_features


def get_poincare_plot_features(nn_intervals: List[float], Plot: bool=False) -> dict:
 
    diff_nn_intervals = np.diff(nn_intervals, axis=0)
    # measures the width of poincare cloud
    sd1 = np.sqrt(np.std(diff_nn_intervals, ddof=1) ** 2 * 0.5)
    # measures the length of the poincare cloud
    sd2 = np.sqrt(2 * np.std(nn_intervals, ddof=1) ** 2 - 0.5 * np.std(diff_nn_intervals, ddof=1) ** 2)
    ratio_sd2_sd1 = sd2 / sd1

    poincare_plot_features = {
        'sd1': sd1,
        'sd2': sd2,
        'ratio_sd2_sd1': ratio_sd2_sd1
    }
    
    return poincare_plot_features


def Features(sample_path):
        ann = wfdb.rdann(sample_path,'atr')   
        rs = ann.sample
        fs = ann.fs

        nnis = np.diff(rs, axis=0)*1000/fs 
          
        # Features extraction based on RR intervals  
        time_domain_features = get_time_domain_features(nnis, fs)
        tdfList = list(time_domain_features.values())
        
        frequency_domain_features = get_frequency_domain_features(nnis)
        fdfList = list(frequency_domain_features.values())
        
        freq, psd = _get_freq_psd_from_nn_intervals(nnis)
        features_from_psd = _get_features_from_psd(freq, psd)
        ffpList = list(features_from_psd.values())
        
        csi_cvi_features = get_csi_cvi_features(nnis)
        ccfList = list(csi_cvi_features.values())
        
        poincare_plot_features = get_poincare_plot_features(nnis, Plot=False)
        ppfList = list(poincare_plot_features.values())
        # print(sample +' Done~')
        
        featureList = tdfList + fdfList + ffpList + ccfList + ppfList
        
        return featureList